import torch
import torch.nn as nn
from torch.nn import functional as F
from metanets import MetaOptimizer, MetaInitializer
from networks import L2normLoss, SmoothL1Loss
import numpy as np
import config
from utils import compute_filter_size, generate_anchors, adaptable_conv2d


class OnlineROAM(nn.Module):

    def __init__(self, feat_extractor):
        super(OnlineROAM, self).__init__()
        self.feat_extractor = feat_extractor
        self.meta_init = MetaInitializer()
        self.meta_opti = MetaOptimizer()
        self.l2loss = L2normLoss()
        self.smoothl1loss = SmoothL1Loss()
        self.meta_opti.eval()

    def compute_resize_feature(self, patch, win_cell_sz):
        with torch.no_grad():
            feat = self.feat_extractor(patch)
            feat = F.interpolate(feat, (win_cell_sz, win_cell_sz), mode='bilinear', align_corners=True)
        return feat

    def initialize(self, feat, label_map, gt_bbox):

        # build initial target classifier
        init_params, init_lrs = self.meta_init.initialize()
        self.meta_opti.initialize(init_lrs)
        target_cell_sz = torch.ceil(gt_bbox[:, 0, 2:]/config.cell_sz)
        filter_size = compute_filter_size(target_cell_sz)
        base_anchor_sizes = filter_size/config.filter_scale*config.cell_sz
        self.map_size = feat.shape[3]
        self.anchors = generate_anchors(self.map_size, base_anchor_sizes)

        # calculate initial update loss
        pred_map, pred_bbox = adaptable_conv2d(feat, init_params, filter_size)
        map_loss = self.l2loss(pred_map, label_map)
        bbox_loss = self.smoothl1loss(pred_bbox, gt_bbox, self.anchors)
        loss = map_loss + bbox_loss
        # meta update
        grads = torch.autograd.grad(loss, init_params)
        self.updated_params = init_params - init_lrs * grads[0]
        pred_map, pred_bbox = adaptable_conv2d(feat, self.updated_params, filter_size)
        lh_map_loss = self.l2loss(pred_map, label_map)
        lh_bbox_loss = self.smoothl1loss(pred_bbox, gt_bbox, self.anchors)
        lh_loss = lh_map_loss + lh_bbox_loss
        print(map_loss.data.item(), lh_map_loss.data.item())
        print(bbox_loss.data.item(), lh_bbox_loss.data.item())
        print(loss.data.item(), lh_loss.data.item())

        self.updating_feats = []
        self.updating_maps = []
        self.updating_bboxes = []
        self.filter_size = filter_size
        self.ref_score = np.max(pred_map.data.cpu().numpy())

    def save_training_samples(self, updating_feat, updating_map, updating_bbox):

        if len(self.updating_feats) == config.max_db_size:
            self.updating_feats.pop(0)
            self.updating_maps.pop(0)
            self.updating_bboxes.pop(0)
        self.updating_feats.append(updating_feat)
        self.updating_maps.append(updating_map)
        self.updating_bboxes.append(updating_bbox)

    def adapt_filter_size(self, target_sz):

        # adapt filter to new size
        target_sz = torch.from_numpy(target_sz[None]).float()
        if torch.cuda.is_available():
            target_sz = target_sz.cuda()
        filter_size = compute_filter_size(target_sz)
        if (self.filter_size != filter_size).any():
            self.filter_size = filter_size
            base_anchor_sizes = self.filter_size / config.filter_scale * config.cell_sz
            # self.anchors = generate_anchors(self.map_size, base_anchor_sizes)
            self.anchors[:, 2:] = base_anchor_sizes

    def meta_update(self):

        # update model
        updating_feats = torch.stack(self.updating_feats, 1)
        updating_maps = torch.stack(self.updating_maps, 1)
        updating_bboxes = torch.stack(self.updating_bboxes, 1)
        n_total = config.n_online_updates * config.n_update_batch
        n_cur_db = updating_feats.size(1)
        all_indices = [i%n_cur_db for i in range(n_total)]
        # all_indices = np.random.permutation(all_indices)
        indices = [all_indices[i*config.n_update_batch:(i+1)*config.n_update_batch]
                   for i in range(config.n_online_updates)]

        for i in range(config.n_online_updates):
            pred_map, pred_bbox = adaptable_conv2d(updating_feats[:, indices[i]], self.updated_params, self.filter_size)
            map_loss = self.l2loss(pred_map, updating_maps[:, indices[i]])
            bbox_loss = self.smoothl1loss(pred_bbox, updating_bboxes[:, indices[i]], self.anchors)
            loss = map_loss + bbox_loss
            grads = torch.autograd.grad(loss, self.updated_params)
            self.updated_params = self.meta_opti.meta_update(self.updated_params, loss, grads[0])

    def localize(self, feat):
        pred_map, pred_bbox = adaptable_conv2d(feat, self.updated_params, self.filter_size)
        return pred_map, pred_bbox